--
-- dice Lua plugin for TS3
-- written 2015 by silmar.alech(at)wombaz.de
--
-- Changes:
-- 2016-07-30 - added plugin announcement on server connect

-- array with clientIDs which also have the plugin installed
local sender = {}

-- list of timestamps to prevent flooding channels with huge responses
local responsetimer = {}

-- log something to the console window (start "ts3client_win64.exe -console")
local function log(msg)
  print("debug: myID="..ts3.getClientID(ts3.getCurrentServerConnectionHandlerID()).." "..msg)
end


-- check message rate
-- return true, if not more than <limit> events within <window> seconds
local function CheckRate(keyword, window, limit)
  local idx = ts3.getCurrentServerConnectionHandlerID()..":"..keyword
  local last = responsetimer[idx] or { timestamp=0, rate=0 }
  local now = os.clock()
  if last.timestamp + window < now then
    last.timestamp = now
    last.rate = 0
  end
  local ratenew = (last.rate + 1) / (now - last.timestamp + window) * window
  if ratenew > limit then
    log("CheckRate over limit keyword="..keyword.." window="..window.." limit="..limit.." current rate="..ratenew)
    return false
  end
  log("CheckRate within limit keyword="..keyword.." window="..window.." limit="..limit.." current rate="..ratenew)

  responsetimer[idx] = { timestamp=now, rate=ratenew }
  return true
end


-- formatted output of dice results
local function CreateResponse(mode, num, dice, value)
--  f1 = counter (1., 2., ...)
--  f2 = dice incl. +- sign for dice throws
--  f3 = equal sign "="
--  f4 = result of a dice throw in bold
--  mode 1: plain number. Only display value in f4 with +- sign
--  mode 2: total. Display "Total:" in f1 and value in f4 with - sign if negative
--  mode 3: dice roll without number. Display sign+dice in f2, value in f4 with - sign if negative
--  mode 4: dice roll with number. Display number in f1, sign+dice in f2, value in f4 with - sign if negative
  local f1, f2, f3, f4 = "     ", "     ", "   ", string.format("%4d", value)

  -- sign is - or space
  local sign = string.sub(string.format("% d", value), 1, 1)

  if mode == 1 then
    f4 = string.format("%+4d", value)
  elseif mode == 2 then
    f1 = "Total:    "
    f2 = ""
  elseif mode == 3 then
    f2 = sign..string.format("%4s", dice)
    f3 = " = "
  elseif mode == 4 then
    f1 = string.format("%3d", num)..". "
    f2 = sign..string.format("%4s", dice)
    f3 = " = "
  end

  local s = "\n  "..f1..f2..f3.."[b]"..f4.."[/b]"  -- result
  s = s:gsub("-", "–") -- replace minus with longer hypen
  return string.gsub(s:gsub("   ", "    "), "       ", "        ") -- expand spaces so things look more aligned
end


local function help(serverConnectionHandlerID)

  if not CheckRate("help", 10, 1) then return nil end

  return "\nWürfelplugin v2 von Silmar Alech\n"..
         "http://www.wombaz.de/uploads/2015/09/lua-dice-plugin-2.zip\n"..
         "(?)                      diese Hilfe\n"..
         "(3d6+1+2d4-3) wirf w6 und w4 mehrfach und addiere/subtrahiere\n"..
         "                          beliebige Anzahl von Seiten und Wurfanzahl möglich\n"..
         "(m)                     wirf eine Münze, auch: (c)\n"..
         "(gm)                   öffne privaten Chat für geheime Würfe (als Spielleiter)\n"..
         "(l)                       liste alle Leute im Channel in zufälliger Reihenfolge"
end


-- create a randomized list of all clients in the current channel
local function RandomizeClients(serverConnectionHandlerID)

  if not CheckRate("list", 10, 1) then return nil end

  local clientID = ts3.getClientID(serverConnectionHandlerID)
  local channelID = ts3.getChannelOfClient(serverConnectionHandlerID, clientID)
  local clients = ts3.getChannelClientList(serverConnectionHandlerID, channelID)
  if #clients > 20 then
    return "Sorry, keine Liste. Zuviele Leute im Channel ("..#clients..")"
  end

  local s = ""
  for i = #clients, 1, -1 do
    local j = math.random(i)
    clients[i], clients[j] = clients[j], clients[i]
    s = "\n"..string.format("%3d", i)..". "..ts3.getClientDisplayName(serverConnectionHandlerID, clients[i])..s
  end

  return s
end


-- throw a coin
local function ThrowCoin(serverConnectionHandlerID, fromName)

  if not CheckRate("coin", 5, 5) then return nil end
  local roll = math.random(2)
  return "\n[b][color=darkred]"..fromName.."[/color][/b] wirft eine Münze... es ist: [b]"..
         (math.random(2) == 2 and "Kopf" or "Zahl") .."[/b]"
end


-- process dice expression
local function RollDice(serverConnectionHandlerID, msg, fromName)
  -- a dice expression is a concatenation of terms joined by + or -
  -- depending on the sign, add/substract the value of the following term
  -- a term can be:
  --   2               add/substract a constant number
  --  d6               throw d6 1 times and add/substract the number
  --  3d20             throw d20 3 times and add/substract the numbers

  -- collect the dice expression into expr and the results into response
  log("RollDice: msg="..msg)
  if not CheckRate("dice", 5, 5) then return nil end

  local expr, response, sum, terms = "", "", 0, 0
  for modifier, value, kw, sides in msg:gmatch("([%+%-]?)(%d*)([dw]?)(%d*)") do
    log("modifier="..modifier.." value="..value.." kw="..kw.." sides="..sides)

    -- create a +1 or -1 depending on the expression sign
    modifier = tonumber(modifier.."1")

    -- add sign to expr as - or +
    expr = expr..string.sub(string.format("%+d", modifier), 1, 1)

    if kw:len() == 0 then

      -- roll no dice, only add/substract a constant
      value = tonumber(value) or 0
      if value ~= 0 then
        expr = expr..value
        value = modifier * value
        response = response..CreateResponse(1, nil, nil, value)
        terms = terms + 1
        sum = sum + value
      end

    else

      -- roll the dice
      value = tonumber(value) or 1
      sides = tonumber(sides) or 6

      if value == 0 or sides < 2 or value > 1000 then

        expr, response = "Schrott", "\nnicht dein Ernst, oder?"
        terms = 1
        break

      else

        -- add multiplicator to expr, if greater than 1
        if value > 1 then expr = expr..value end

        -- add dN to expr
        local dice = kw..sides
        expr = expr..dice

        for i = 1, value do
          local roll = modifier * math.random(sides)
          if value == 1 then
            response = response..CreateResponse(3, nil, dice, roll)
          -- don't display each roll if more than 25 rolls requested
          elseif value <= 25 then
            response = response..CreateResponse(4,   i, dice, roll)
          end
          terms = terms + 1
          sum = sum + roll
        end
      end
    end
  end

  -- return empty response if no terms were processed
  if terms == 0 then return nil end

  -- strip leading and trailing + from expr
  expr = expr:match("^%+?(.-)%+?$")

  -- construct final response
  response = "\n[b][color=red]"..fromName.."[/color][/b] würfelt [b]("..expr..")[/b]:"..response

  -- add "Total:" on complex expressions
  if terms > 1 then
    response = response..CreateResponse(2, nil, nil, sum).."     ("..expr..")"
  end

  log("expr="..expr)
  return response
end


-- announce the presence of the plugin to the other clients in the current channel
-- targetMode: 0=PluginCommandTarget_CURRENT_CHANNEL,  PluginCommandTarget_CLIENT=2 (list in clientids)
local function announceDicePlugin(serverConnectionHandlerID, targetMode, clients)
  local myID = ts3.getClientID(serverConnectionHandlerID)
  -- the list of clients must be terminated by id=0
  if targetMode == 2 then table.insert(clients, 0) end
  log("announceDicePlugin targetMode="..targetMode.." clients=("..table.concat(clients, ",")..")")
  ts3.sendPluginCommand(serverConnectionHandlerID, "dice "..myID.." "..targetMode, targetMode, clients)
end


-- if I moved in or out of a channel, announce the presence of the plugin to the other clients in the channel
local function onClientMoveEvent(serverConnectionHandlerID, clientID, oldChannelID, newChannelID, visibility, moveMessage)
  if clientID == ts3.getClientID(serverConnectionHandlerID) then
    announceDicePlugin(serverConnectionHandlerID, 0, {})
  end
end


-- if I was moved in or out of a channel, announce the presence of the plugin to the other clients in the channel
local function onClientMoveMovedEvent(serverConnectionHandlerID, clientID, oldChannelID, newChannelID, visibility, moverID, moverName, moverUniqueIdentifier, moveMessage)
  if clientID == ts3.getClientID(serverConnectionHandlerID) then
    announceDicePlugin(serverConnectionHandlerID, 0, {})
  end
end

-- on connection, announce the presence of the plugin to the other clients in the channel
local function onConnectStatusChangeEvent(serverConnectionHandlerID, status, errorNumber)
  log("onConnectStatusChangeEvent serverConnectionHandlerID="..serverConnectionHandlerID.." status="..status)

  if status == ts3defs.ConnectStatus.STATUS_CONNECTION_ESTABLISHED then
    announceDicePlugin(serverConnectionHandlerID, 0, {})
  end
end

-- listen to announcements of other clients with the dice plugin
-- if an announcement is detected, save the clientID into sender{}
-- and notify sender as well if the announcement was a channel broadcast
local function onPluginCommandEvent(serverConnectionHandlerID, pluginName, pluginCommand)
  log("onPluginCommandEvent pluginName="..pluginName.." pluginCommand="..pluginCommand)
  local clientID, targetMode = pluginCommand:match("dice (%d+) (%d+)")
  if pluginName ~= "lua_plugin" or clientID == nil then return end
  clientID, targetMode = tonumber(clientID), tonumber(targetMode)
  local idx = serverConnectionHandlerID..":"..clientID
  local is_unknown_clientID = sender[idx] == nil
  sender[idx] = clientID
  log("register sender "..idx)

  -- notify remote if targetmode was channel broadcast
  if is_unknown_clientID and targetMode == 0 and clientID ~= ts3.getClientID(serverConnectionHandlerID) then
    log("announce back to clientID="..clientID)
    announceDicePlugin(serverConnectionHandlerID, 2, { clientID })
  end

end


-- determine sender
-- sender is the lowest clientID from clients which is also in sender{}
local function GetSenderID(serverConnectionHandlerID, clients)

  table.sort(clients)
  for i, v in ipairs(clients) do
    if sender[serverConnectionHandlerID..":"..v] ~= nil then return v end
  end
  return nil
end


-- onTextMessageEvent gets called upon arrival of a text message
local function onTextMessageEvent(serverConnectionHandlerID, targetMode, toID, fromID, fromName, fromUniqueIdentifier, message, ffIgnored)
  local response

  if message:sub(1, 5) == "debug" then return 0 end

  -- turn to lower case and strip all whitespace
  local msg = string.gsub(message:lower(), "%s+", "")

  local myID = ts3.getClientID(serverConnectionHandlerID)
  if msg:match("^%([mc]%)$") then
    response = ThrowCoin(serverConnectionHandlerID, fromName)
  elseif msg:match("^%(gm%)$") then
    if fromID == myID then
      log("redirect gm message to myself msg="..msg)
      targetMode = ts3defs.TextMessageTargetMode.TextMessageTarget_CLIENT
      toID = fromID
      fromName = "Spielleiter"
      response = "Privater Würfelchannel"
    elseif sender[serverConnectionHandlerID..":"..fromID] == nil then
      response = "Du mußt das Würfelplugin selbst installiert haben, um mit dir selbst würfeln zu können!"
    end
  elseif msg:match("^%(%?%)$") then
    response = help(serverConnectionHandlerID)
  elseif msg:match("^%(l%)$") then
    log("matched channel list request")
    response = RandomizeClients(serverConnectionHandlerID)
  else
    -- match complete dice expression and strip (). msg is empty on invalid expressions.
    response = RollDice(serverConnectionHandlerID, msg:match("^%(([dw%d%+%-]+)%)$"), fromName)
  end

  log("response="..(response or "(nil)"))
  if response == nil then return 0 end

  -- we have a response, send it if we are the sender
  if targetMode == ts3defs.TextMessageTargetMode.TextMessageTarget_CHANNEL then

    -- request was received in a channel, so send response to that channel
    local channelID = ts3.getChannelOfClient(serverConnectionHandlerID, fromID)
    local clients = ts3.getChannelClientList(serverConnectionHandlerID, channelID)
    local senderID = GetSenderID(serverConnectionHandlerID, clients)
    log("received channel chat fromID="..fromID.." toID="..toID.." senderID="..(senderID or "(nil)"))

    if senderID == myID then
      ts3.requestSendChannelTextMsg(serverConnectionHandlerID, response, channelID)
    elseif senderID == nil then
      -- we got no sender, so announce the dice plugin to get a sender
      announceDicePlugin(serverConnectionHandlerID, 0, {})
    end

  elseif targetMode == ts3defs.TextMessageTargetMode.TextMessageTarget_CLIENT then

    -- request was received in private chat, so send response to that chat partner
    local senderID = GetSenderID(serverConnectionHandlerID, {toID, fromID})
    log("received private chat fromID="..fromID.." toID="..toID.." senderID="..(senderID or "(nil)"))

    if senderID == myID then
      ts3.requestSendPrivateTextMsg(serverConnectionHandlerID, response, (myID == fromID) and toID or fromID)
    elseif senderID == nil then
      -- we got no sender, so announce the dice plugin to get a sender
      announceDicePlugin(serverConnectionHandlerID, 2, {toID, fromID})
    end

  end

  return 0
end


diceroll_events = {
  onTextMessageEvent = onTextMessageEvent,
  onPluginCommandEvent = onPluginCommandEvent,
  onClientMoveEvent = onClientMoveEvent,
  onClientMoveMovedEvent = onClientMoveMovedEvent,
  onConnectStatusChangeEvent = onConnectStatusChangeEvent
}
